define([
    "aps/Container",
    "aps/i18n",
    "aps/Message",
    "aps/ResourceStore",
    "aps/Select",
    "dijit/registry",
    "dojo/_base/declare",
    "dojo/Deferred"
    ], function(
        Container,
        _,
        Message,
        Store,
        Select,
        registry,
        declare,
        Deferred
    ) {
    return declare([Container], {
        // Store of phone number blocks. Optional, if property is null that default store will be used
        phoneNumberBlocksStore: null,

        // Store of phone numbers. Optional, if property is null that default store will be used
        phoneNumbersStore: null,

        // Orgainization ID. Property is reqiured in case default store of phone number blocks is used
        organizationId: null,

        // Owner of phone number. Optional, it allows to get phone numbers of owner in addition to the free numbers
        phoneNumberOwner: null,

        // Value to bind phone number ID with model through 'at' operation. Reqiured.
        // example: new PhoneNumberSelector({value: at('model.phoneNumber.aps', 'id')})
        value: null,

        // Label. Optional
        label: _("Phone Number"),

        // Flag to indicate that no available phone numbers. Output property, don't set it manualy
        noAvailableNumber: true,

        // Minimum limit of phone numbers in block to add this block in PhoneNumberSelector. Optional
        minAvailableLimit: 1,

        // Flag to indicate that PhoneNumberSelector display only block list without phone numbers. Optional
        showBlocksOnly: false,

        // Flag to indicate that PhoneNumberSelector can be set in None (without any phone number). Optional
        allowNone: false,

        // TODO: remove after add checking of 'phoneNumberOwner' and 'value'
        // Flag to indicate that PhoneNumberSelector can mark initial value as "Current". Optional
        initialValueIsCurrent: true,

        // waitForLoad: [readonly] Deferred
        //  Deferred action to wait for phone selector is loaded. Output property, don't set it manualy.
        //  Result: hasNumbers - boolean value indicating whether there're available phone numbers.
        waitForLoad: new Deferred(),

        postCreate: function () {
            var self = this;
            self.inherited(arguments);

            if(self.phoneNumberBlocksStore === null) {
                if (self.organizationId === null) return;

                self.phoneNumberBlocksStore = new Store({
                    apsType: "http://www.parallels.com/LyncHPv2/phoneNumberBlock/1.1",
                    target: "/aps/2/resources/" + self.organizationId + "/blocks/?sort(+name)"
                });
            }

            if (self.phoneNumbersStore === null) {
                var ownerFilter = self.phoneNumberOwner === null ?
                    "isnull(ownerSip)" :
                    "or(isnull(ownerSip),eq(ownerSip,sip:" + self.phoneNumberOwner.replace(/^sip\:/, '') + "))";

                self.phoneNumbersStore = new Store({
                    apsType: "http://www.parallels.com/LyncHPv2/phoneNumber/1.0",
                    target: "/aps/2/resources/?" + ownerFilter + ",sort(+number)"
                });
            }

            var buildControlsOnBlocks = function(phoneNumbers, blocks) {
                var blocksMap = {};
                var i;
                for (i = 0; i<blocks.length; i++) {
                    blocksMap[blocks[i].aps.id] = blocks[i];
                    blocks[i].phoneNumbers = [];
                }
                var phoneNumber;
                for (i = 0; i<phoneNumbers.length; i++) {
                    phoneNumber = phoneNumbers[i];
                    if (phoneNumber.blockId in blocksMap) {
                        blocksMap[phoneNumber.blockId].phoneNumbers.push(phoneNumber);
                    }
                }

                function isCurrentBlock(block) {
                    for (var i=0; i<block.phoneNumbers.length; i++){
                        var blockNumber = block.phoneNumbers[i];
                        if ((self.value && blockNumber.aps.id == self.value) ||
                                    (self.phoneNumberOwner && ("sip:" + self.phoneNumberOwner.replace(/^sip\:/, '')) == blockNumber.ownerSip)) {
                            return true;
                        }
                    }
                    return false;
                }

                function getAvailableBlocksCount(blocks) {
                    var result = 0;
                    for (var i=0; i<blocks.length; i++) {
                        if (blocks[i].phoneNumbers.length > 0) {
                            result++;
                        }
                    }
                    return result;
                }

                var phoneNumberBlockSelectOptions = [];
                var phoneNumberSelectOptions = [];
                var multipleBlocksAvailable = self.showBlocksOnly ? true : getAvailableBlocksCount(blocks) > 1;
                if (multipleBlocksAvailable) {
                    if (self.allowNone) {
                        phoneNumberBlockSelectOptions.push({label: _("None"), value: ""});
                    }
                    phoneNumberBlockSelectOptions.push({label: _("--- Select Phone Number Block ---"), value: "", disabled: true});
                }

                // TODO: added checking: current phoneNumber.Owner == 'phoneNumberOwner' and 'value' == null
                var initValueEmpty = self.value === null && !self.phoneNumberOwner;
                for (i = 0; i<blocks.length; i++) {
                    var block = blocks[i];
                    if (block.phoneNumbers.length >= self.minAvailableLimit) {

                        self.noAvailableNumber = false;

                        var numbers = [];
                        if(!self.showBlocksOnly){
                            if (self.allowNone && !multipleBlocksAvailable) {
                                numbers.push({label: _("None"), value: ""});
                            }
                            numbers.push({label: _("--- Select Phone Number ---"), value: "", disabled: true});

                            for (var j=0; j<block.phoneNumbers.length; j++) {
                                phoneNumber = block.phoneNumbers[j];
                                phoneNumber.block = block;
                                var isSelected = (self.value && phoneNumber.aps.id == self.value) ||
                                    (self.phoneNumberOwner && ("sip:" + self.phoneNumberOwner.replace(/^sip\:/, '')) == phoneNumber.ownerSip);
                                numbers.push({
                                    label: (isSelected && self.initialValueIsCurrent) ? _("__phoneNumber__ (Current)", { phoneNumber: phoneNumber.displayNumber }) : phoneNumber.displayNumber,
                                    value: phoneNumber.aps.id,
                                    selected: isSelected,
                                    model: phoneNumber
                                });
                                if (self.value === null || isSelected) {
                                    self.value = phoneNumber.aps.id;
                                    phoneNumberSelectOptions = numbers;
                                }
                            }
                        }

                        var isSelectedBlock = isCurrentBlock(block);
                        phoneNumberBlockSelectOptions.push({
                            label: (isSelectedBlock && !initValueEmpty && self.initialValueIsCurrent) ? _("__phoneBlock__ (Current)", { phoneBlock: block.name }) : block.name,
                            value: block.aps.id,
                            selected: isSelectedBlock,
                            model: { data: block, numbers: numbers }
                        });
                    }
                }
                if(self.noAvailableNumber) {
                    phoneNumberBlockSelectOptions = [];
                    phoneNumberSelectOptions = [];
                    phoneNumberBlockSelectOptions.push({ label: self.showBlocksOnly ? _("No available phone number blocks") : _("No available phone numbers"), value: "", disabled: true, selected: true });
                }

                self.addChild(new Select({
                    id: self.id + "_bocks",
                    options: phoneNumberBlockSelectOptions,
                    visible: multipleBlocksAvailable || self.noAvailableNumber,
                    onChange: function(value){
                        if (self.showBlocksOnly) return;

                        var block = registry.byId(self.id + "_bocks").getOption(value)[0].model;
                        var phoneNumberSelector = registry.byId(self.id + "_numbers");
                        if (block !== null && block !== undefined) {
                            phoneNumberSelector.set("options", block.numbers);
                            phoneNumberSelector.set("visible", true);
                            phoneNumberSelector.set("value", phoneNumberSelector.focusNode.value);
                        }
                        else {
                            phoneNumberSelector.set("options", []);
                            phoneNumberSelector.set("visible", false);
                            phoneNumberSelector.set("value", "");
                        }
                        phoneNumberSelector.startup();
                    },
                    "style": "margin-right: 10px;"
                }));

                var phoneNumberSelect = new Select({
                    id: self.id + "_numbers",
                    value: self._refs.value,
                    options: phoneNumberSelectOptions,
                    visible: !self.showBlocksOnly && phoneNumberSelectOptions.length > 0
                });

                self.addChild(phoneNumberSelect);

                if (self.value !== null) {
                    phoneNumberSelect.set("value", self.value);
                }

                self.waitForLoad.resolve(!self.noAvailableNumber);
            };

            var buildControlsOnNumbers = function(phoneNumbers) {
                var blocks = self.phoneNumberBlocksStore.query();
                if (blocks instanceof Array) {
                    buildControlsOnBlocks(phoneNumbers, blocks);
                } else {
                    blocks.then(function(blocks) {
                        buildControlsOnBlocks(phoneNumbers, blocks);
                    });
                }
            };

            var phoneNumbers = self.phoneNumbersStore.query();
            if (phoneNumbers instanceof Array) {
                buildControlsOnNumbers(phoneNumbers);
            } else {
                phoneNumbers.then(buildControlsOnNumbers);
            }
        },
        getNumber: function(id) {
            return registry.byId(this.id + "_numbers").getOption(id)[0].model;
        },
        getBlock: function(id) {
            var blockSelector = registry.byId(this.id + "_bocks");
            if (id === "" || id === null || id === undefined){
                id = blockSelector.get("value");
            }
            if (id === "") {
                return null;
            }
            return blockSelector.getOption(id)[0].model.data;
        }
    });
});