define([
	"dojo/_base/declare",
	"dojo/store/Memory",
	"dojo/_base/lang",
	"dojo/store/util/QueryResults",
	"dojo/_base/array",
	"./rql/js-array.js",
	"dojo/store/util/SimpleQueryEngine"
], function(declare, Memory, lang, queryResults, arr, jsArray, simpleQueryEngine){
	return declare(Memory, {
		// summary:
		//		This is a basic in-memory object store. It implements dojo.store.api.Store.
		// queryEngine: Function
		//		Defines the query engine to use for querying the data store
		queryEngine: jsArray.query,
		getIdentity: function(object){
			// summary:
			//		Returns an object's identity
			// object: Object
			//		The object to get the identity from
			// returns: Number
			//		value idProperty
			return lang.getObject(this.idProperty, false, object);
		},
		put: function(object, options){
			// 	summary:
			//		Stores an object
			// 	object: Object
			//		The object to store.
			// 	options: Store.PutDirectives??
			//		Additional metadata for storing the data.  Includes an "id"
			//		property if a specific id is to be used.
			// returns: Number
			var data = this.data,
				index = this.index,
				id = this.getIdentity(object);
			if(id in index){
				// object exists
				if(options && options.overwrite === false){
					throw new Error("Object already exists");
				}
				// replace the entry in data
				data[index[id]] = object;
			} else {
				// add the new object
				index[id] = data.push(object) - 1;
			}
			return id;
		},
		setData: function(data){
			// summary:
			//		Sets the given data as the source for this store, and indexes it
			// data: Object[]
			//		An array of objects to use as the source of data.
			if(data.items){
				// just for convenience with the data format IFRS expects
				this.idProperty = data.identifier;
				data = this.data = data.items;
			}else{
				this.data = data;
			}
			this.index = {};
			for(var i = 0, l = data.length; i < l; i++){
				this.index[this.getIdentity(data[i])] = i;
			}
		},
		_objectToQuery: function(que, options){
			options = options || {};

			var filters	 = [];

			if(options.sort)
				arr.forEach(options.sort, function(item){
					filters.push("sort(" + (item.descending ? "-" : "+") + item.attribute +")");
				});
			if(options.type)
				filters.push("implementing(" + options.type + ")");
			if(que){
				if(que && typeof que == "string"){
					filters[0] = que;
				}
				else
					for(var key in que){
						if(que.hasOwnProperty(key)) {
							if(typeof que[key] === "boolean"){
								filters.push("eq(" + key + "," + que[key] + "()" + ")");
								continue;
							}
							filters.push((que[key].indexOf && que[key].indexOf('*') > -1 ? "like(" : "eq(") +
								key + "," + encodeURIComponent(que[key]) + ")");
						}
					}
			}

			return filters.length > 0 ? (filters.length == 1 ? filters[0] : ("and(" + filters.join(",") + ")")) : "";
		},
		query: function(query, options){
			// summary:
			//		Queries the store for objects.
			// query: Object
			//		The query to use for retrieving objects from the store.
			// options: dojo/store/api/Store.QueryOptions?
			//		The optional arguments to apply to the resultset.
			// returns: dojo/store/api/Store.QueryResults
			//		The results of the query, extended with iterative methods.
			//
			// example:
			//		Given the following store:
			//
			// 	|	var store = new Memory({
			// 	|		data: [
			// 	|			{id: 1, name: "one", prime: false },
			//	|			{id: 2, name: "two", even: true, prime: true},
			//	|			{id: 3, name: "three", prime: true},
			//	|			{id: 4, name: "four", even: true, prime: false},
			//	|			{id: 5, name: "five", prime: true}
			//	|		]
			//	|	});
			//
			//	...find all items where "prime" is true:
			//
			//	|	var results = store.query({ prime: true });
			//
			//	...or find all items where "even" is true:
			//
			//	|	var results = store.query({ even: true });
			options  = options || {};
			query	 = query || {};

			if (options && options.apsType)
				options.type = encodeURIComponent(options.apsType);

			//RQL can not work with arbitrary RegExp. Need for aps/ComboBox.
			for (var key in query) {
				if (query.hasOwnProperty(key))
				if(query[key] && query[key].test)
					return queryResults(simpleQueryEngine(query, options)(this.data));
			}

			var qe = this.queryEngine(this._objectToQuery(query, options)),
				results  = qe(this.data);

			if(options.count && options.count != Infinity){
				var oldResults = results;
				results = results.slice(options.start, options.start + options.count);
				results.total = oldResults.length;
			} else {
				results.total = results.length;
			}

			return queryResults(results);
		}
	});
});